/*
 * ERIKA Enterprise Basic - a tiny RTOS for small microcontrollers
 *
 * Copyright (C) 2002-2007  Evidence Srl
 *
 * This file is part of ERIKA Enterprise Basic.
 *
 * ERIKA Enterprise Basic is free software; you can redistribute it
 * and/or modify it under the terms of the GNU General Public License
 * version 2 as published by the Free Software Foundation.
 *
 * ERIKA Enterprise Basic is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License version 2 for more details.
 *
 * You should have received a copy of the GNU General Public License
 * version 2 along with ERIKA Enterprise Basic; if not, write to the
 * Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA 02110-1301 USA.
 */

/*
 * Author: Gianluca Franchino.
 * Affiliation: Retis Lab. Scuola Superiore Sant'Anna. Pisa (Italy).
 * Contacts: g.franchino@sssup.it; gianluca@evidence.eu.com
 * Date: 02/27/2008.
 * 
 */

#include <stdlib.h>
#include <stdio.h> 
#include "ee.h"
#include "cpu/pic30/inc/ee_irqstub.h"
#include "ee_internal.h"

#include "eeuart.h"
#include "code.h"

// Primary (XT, HS, EC) Oscillator without PLL
_FOSCSEL(FNOSC_PRIPLL);
// OSC2 Pin Function: OSC2 is Clock Output - Primary Oscillator Mode: XT Crystanl
_FOSC(OSCIOFNC_ON & POSCMD_XT);
// Watchdog Timer Enabled/disabled by user software
_FWDT(FWDTEN_OFF);
// Disable Code Protection
_FGS(GCP_OFF);            							


/* Program the Timer1 peripheral to raise interrupts */
void T1_program(void)
{
	T1CON = 0;		/* Stops the Timer1 and reset control reg	*/
	TMR1  = 0;		/* Clear contents of the timer register	*/
	PR1   = 0x9c40;		/* Load the Period register wit the value 0x9c40 (1ms)*/
	IPC0bits.T1IP = 5;	/* Set Timer1 priority to 5		*/
	IFS0bits.T1IF = 0;	/* Clear the Timer1 interrupt status flag	*/
	IEC0bits.T1IE = 1;	/* Enable Timer1 interrupts		*/
	T1CONbits.TON = 1;	/* Start Timer1 with prescaler settings at 1:1
				* and clock source set to the internal 
				* instruction cycle			*/
}

/* Clear the Timer1 interrupt status flag */
void T1_clear(void)
{
	IFS0bits.T1IF = 0;
}

/* 
 * Program the INT0 (or INT4) pin to raise an interrupt everytime  
 * there is a rising edge (i.e. everytime FIFOP goes high).
 */   
#ifndef FLEX_DEMO_BOARD  
void INT0_config(void)
{
	IFS0bits.INT0IF = 0;/* clear the interrupt flag */
	IPC0bits.INT0IP = 0b001; /* set int0 priority to 1 */
	INTCON2bits.INT0EP = 0; /* set interrupt on rising edge */
	IEC0bits.INT0IE = 1; /* enable the interrupt */
}
#else
void INT4_config(void)
{
	IFS3bits.INT4IF = 0;/* clear the interrupt flag */
	IPC13bits.INT4IP = 0b001; /* set int0 priority to 1 */
	INTCON2bits.INT4EP = 0; /* set interrupt on rising edge */
	IEC3bits.INT4IE = 1; /* enable the interrupt */
}

#endif

/*
 * This function sets the hardware, namely, the I/O pins, the UART port,
 * the SPI port and the EasyBee/CC2420.
 */ 
void hardware_init()
{
	/*set the I/O pins as all digital */
	AD1PCFGH = 0xFFFF;
	AD1PCFGL = 0xFFFF;

    /* Set the I/O pins direction */
	TRIS_RESETn = 0;
	TRIS_VREG_EN = 0;
	TRIS_FIFO = 1;
	TRIS_FIFOP = 1;
	TRIS_CSn = 0;

	
	/*Initialize INT0 linked to FIFOP pin*/
	#ifndef FLEX_DEMO_BOARD 
	INT0_config();
	#else
	INT4_config();
	#endif
	

	hal_console_init(CONSOLE_PORT);
 	hal_console_out_str("\n\n\r+++++++++++++++++++++++++++++++++++++++");
	hal_console_out_str("\n\r+Hello from the FLEX soul!\n\r+Console opened!");

	hal_radio_spi_init(RADIO_SPI2);
	hal_console_out_str("\n\r+Spi port opened!");
	
	if (hal_radio_init() != 0) 
		hal_console_out_str("\n\r+ERROR INITIALIZING THE RADIO!");
	else 
		hal_console_out_str("\n\r+Radio initialized correctly!");
		
	hal_console_out_str("\n\r+++++++++++++++++++++++++++++++++++++++\n\r");
	
	hal_set_channel(0);	
	hal_set_rx();
	
	/*
	 * Set the buffer indexes
	 */
	rx_read_index =  0; 
	rx_write_index = 1;
}

/*
 * This function performs the channel access using a CSMA/CA algorithm. 
 */
int transmit_csma_ca()
{
    BYTE a,b;
    BYTE back_off_cycles;
    
    a = 4; /* number of times to try the channel access */

   /*
    * We get a random value using the timer 1 value
    */  
	back_off_cycles = TMR1; 
	
	/*
	 * In order to perform the carrier sensing operation, the transceiver should
	 * be in receiving mode for at least  8 symbols (8 * 16 usec) before trying
	 * to transmit using the CCA.
	 * So, if the transceiver was not in receiving mode, enable receiving mode
	 * and wait for at least 8 * 16 usec.
	 * Notice that, in this demo the transceiver is already in receiving mode when
	 * TaskSend calls this routine (transmit_csma_ca()).
	 * 
	 * For further information see the CC2420 data sheet.
	 */   

    while (a--)
    {
        // If transmission was started, return.        
        if ( hal_set_tx_on_cca() & 0x08 )
        {
            return 0; /* tx ok */
        }
		
		back_off_cycles <<= 1;

        if ( back_off_cycles == 0x00 || back_off_cycles == 0xFF ) 
			back_off_cycles = TMR1;
			
        /*
         * Wait until backoff cycles expire.
         */
        b = back_off_cycles;
        while (b--);
    }
    return -1; /* tx failed */
}


/* 
 * This is an ISR Type 2 which is attached to the Timer 1 peripheral IRQ pin
 * The ISR simply calls CounterTick to implement the timing reference
 */
ISR2(_T1Interrupt)
{
	/* clear the interrupt source */
	T1_clear();

	/* count the interrupts, waking up expired alarms */
	CounterTick(myCounter);
}

/* 
 * This is an ISR Type 2 which is attached to the INT0 IRQ.
 * The ISR controls the EasyBee FIFOP status and actives TaskAcquire.
 */

#ifndef FLEX_DEMO_BOARD 
ISR2 (_INT0Interrupt)
{
	IFS0bits.INT0IF = 0;

	ActivateTask(TaskAcquire);
		
}
#else
ISR2 (_INT4Interrupt)
{
	IFS3bits.INT4IF = 0;

	ActivateTask(TaskAcquire);
		
}
#endif
/*
 * This Task sends the message "msg" every time it is activated
 */
TASK(TaskSend)
{ 
	/*
	 * msg length = 28
	 */ 
	 
	GetResource(S2); 
	hal_store_msg(msg,28);

	/* if the transceiver is still transmitting
	 * then give up!
	 */ 
	if(hal_radio_get_status() & 0x08) {
		ReleaseResource(S2);
		return;
	}
	
#ifndef USE_CSMA
	hal_set_tx();
#else
	static int csma_fail;
	
	if (transmit_csma_ca() != 0) {
		GetResource(S1);
		hal_console_out_str("\n\rCSMA FAILED: MSG not sent!!!\n\r");
		ReleaseResource(S1);
		csma_fail++;
		/*
		 * if the CSMA fails for more than 5 times probably the transceiver is not working, 
		 * it should be reset and initialized again. 
		 * 
		 */
		if (csma_fail > 5) {
			csma_fail = 0;
			hal_radio_init();
			hal_set_rx();		
		}
	}
	else 
		csma_fail = 0;
	/*	hal_console_out_str("\n\rMSG not sent!");
	*/
#endif
	
	/*
	 * if the rx fifo of the transceiver overflows
	 * flush it
	 */
	if (FIFO == 0 && FIFOP == 1) {
		hal_console_out('F');
		RX_FIFO_FLUSH();
	}
		
	ReleaseResource(S2);	

}

/*
 * This Task retrieve a message from the receiving buffer "rx_buffer" and 
 * sends it to the console.
 * Note that, since it is using time consuming functions like sprintf and 
 * hal_console_out_str, this task takes about 10 msec of execution.
 */

TASK(TaskElaborate)
{
	static unsigned int num = 0;
	char str[25];

	num++;	
	
	GetResource(S1);
	hal_console_out_str("\n\r");
	//hal_console_out_str("                                                  \r");
	ReleaseResource(S1);
	
	sprintf(str,"Msg number %u  -> ",num);
	
	GetResource(S1);	
	hal_console_out_str(str);
	ReleaseResource(S1);
	
	GetResource(S0);
	while (abs(rx_read_index - rx_write_index) > 1 ) {
		rx_read_index = (rx_read_index + 1) % BUFFER_SIZE;
		GetResource(S1);
		hal_console_out(rx_buffer[rx_read_index]);
		ReleaseResource(S1);
	}
	ReleaseResource(S0);
	/*
	GetResource(S1);
	hal_console_out_str("\n\rMSG received!");
	ReleaseResource(S1);
	*/
}

/*
 * This Task is activated by the ISR of INT0, that is, 
 * when EasyBee/CC2420 reports that there is a message on the 
 * RX_FIFO through the FIFOP pin.
 * The task retrieves the message from the RX_FIFO of the tranceiver
 * and memeorizes it in receiveing buffer rx_buffer. 
 */
TASK(TaskAcquire)
{
	WORD length, i;
	
	if (FIFO == 0) {
		
		GetResource(S1); 
		hal_console_out_str("\n\rFIFO OVERFLOW!");
		ReleaseResource(S1);
		GetResource(S2);
		RX_FIFO_FLUSH();
		ReleaseResource(S2);
		return ;	
	}
	GetResource(S2);
	length = hal_get_msg(radio_msg);
	ReleaseResource(S2);

	if (length != -1) {
		/*If CRC is correct go ahead*/
		if (radio_msg[length] & 0x80) {
			/*
			 * The last two bytes in the buffer are the RSSI value 
			 * and the CRC value respectively. They are not part 
			 * of the message (payload), then we can discard them.
			 * For further information see the CC2420 datasheet  
			 */
			length -= 2;
			
			/* if there is no room in the buffer discard the message
			 * ,that is, return without saving the message.
			 */
			 
			GetResource(S0);
			
			if (rx_write_index > rx_read_index) {
				if (BUFFER_SIZE - rx_write_index + rx_read_index
					 - 1 < length) {
				 /*no room in the msg buffer*/
				 
					ReleaseResource(S0);
					return;
					}
			} else if (rx_read_index - rx_write_index - 1 < length) {
					/*no room in the msg buffer*/
					ReleaseResource(S0);
					return;
					}
			/* 
			 * else if there is room store the message in the buffer
			 *
			 * Notice that, i = 1 because 
			 * the first byte in radio_msg[] is the msg length, 
			 * then we discard the first byte.
			 */
			ReleaseResource(S0);
			
			for (i = 1; i <= length ; i++) {
				GetResource(S0);
				rx_buffer[rx_write_index++] = radio_msg[i];
				rx_write_index = rx_write_index % BUFFER_SIZE;
				ReleaseResource(S0); 
			}
			/*
			 * Active the message elaboration Task
			 */
			 	ActivateTask(TaskElaborate);		
		} else {
			 /* CRC failed */
			    /*
				GetResource(S1);
				hal_console_out('C');
				ReleaseResource(S1);
				*/
			}		
	}
	
	
}

int main(void)
{
	/* Clock setup for 40MIPS */
	CLKDIVbits.DOZEN   = 0;
	CLKDIVbits.PLLPRE  = 0;
	CLKDIVbits.PLLPOST = 0;
	PLLFBDbits.PLLDIV  = 78;
	
	/* Wait for PLL to lock */
	while(OSCCONbits.LOCK!=1);
	
	/* Program Timer 1 to raise interrupts */
	T1_program();

	hardware_init();
	/* 
	 * Program a cyclic alarm which will fire after an offset of 10 counter 
	 * ticks, and after that periodically every 100 ticks. 
	 * Please note that 1 tick = 1 ms 
	 */
	SetRelAlarm(AlarmSend, 10, 100); 

	/* Forever loop: background activities */
	for (;;) {
		};
	
	return 0;
}
